/*
 * The original software was
 * cm-cracker: a UNIX password cracker for the Thinking Machines CM/2
 *             and CM/200 Connection Machines.
 *
 * Copyright (C) 1992, UNI-C (Danish Computing Centre for Research and
 * Education) and Michael Glad.
 *
 * This software was ported from UNIX to MS Windows by Andy Baron
 * with the help of Andy Pozo, Midwestern Commerce, Inc.
 *
 * Copyright (C) 1996, Midwestern Commerce, Inc.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <fcntl.h>
#include <ctype.h>

/*
 * typedefs
 */

struct word {
  char *txt;
  struct word *nextw;
};
typedef struct word word;

struct hashent {
  char *txt;
  int created;
  struct hashent *fwrd, *bwrd;
};
typedef struct hashent hashent;

/*
 * Prototypes
 */

void bzero(void* buf, unsigned int len);
char *nextline(char *buf);
char *Malloc(int n);
char *save_str(char *p);
word *read_txt_file(FILE *fp);
word *setup_default_rules(void);
void setup_tables(void);
void bad_rule(char *buf, char c);
void check_char(char *buf, char **pp);
void check_int(char *buf, char **pp);
void check_char_or_class(char *buf, char **pp);
void check_rules(word *rules);
void rcheck(char *word, char *rule);
void interpret_rules(word *rules, word *dictionary, FILE *fp);
void usage(void);
void main(int argc, char **argv);

/*
 * Misc.
 */

char *Malloc(n)
     int n;
{
  char *p;

  if((p = (char*)malloc(n)) == (char*)NULL) {
    fprintf(stderr, "Malloc: ran out of core -- aborting\n");
    exit(1);
  }
  return p;
}

char *save_str(p)
     char *p;
{
  char *y;
  y = (char*)Malloc(strlen(p) + 1);
  strcpy(y, p);
  return y;
}

#define ALLOC(ty) ((struct ty *) Malloc(sizeof(struct ty)))

#define MAXLEN 100

/*
 * Hashtable to do some uniq'ing of generated words.
 * We assume that the input dictionaries reasonably
 * sorted.
 */

#define MAXAGE 250

hashent *hashtable[0xff+1];
hashent *freelist;
int hash_init = 0;
int now, searches, requests, allocs, outwords;

#define unlk(p) \
  (p)->fwrd->bwrd = (p)->bwrd;   \
  (p)->bwrd->fwrd = (p)->fwrd

#define insert_front(l, p) \
 (p)->bwrd = (l); \
 (p)->fwrd = (l)->fwrd; \
 (l)->fwrd->bwrd = (p); \
 (l)->fwrd = (p)

int is_uniq(txt)
     char *txt;
{
  int hash, i;
  hashent *hp, *nexthp, *head;
  char *p1, *p2;

  /*
   * Initialize if not already done
   */

  if(!hash_init) {
    for(i = 0; i < 256; i++) {
      hp = ALLOC(hashent);
      hp->fwrd = hp->bwrd = hp;
      hashtable[i] = hp;
    }
    freelist = ALLOC(hashent);
    freelist->fwrd = freelist->bwrd = freelist;
    hash_init++;
  }

  requests++;

  /*
   * Compute hash value
   */

  hash = txt[0] ^ txt[1] ^ txt[2] ^ txt[3] ^
         txt[4] ^ txt[5] ^ txt[6] ^ txt[7];
  hash &= 0x7f;

  /*
   * Go and search relevant hash queue
   * discarding elements too old
   */

  for(head = hashtable[hash], hp = head->fwrd; hp != head; hp = nexthp) {
    nexthp = hp->fwrd;
    searches++;
    if(now - hp->created > MAXAGE) {
      unlk(hp);
      insert_front(freelist, hp);
      continue;
    }

    if(txt[0] == hp->txt[0] && txt[1] == hp->txt[1] &&
       txt[2] == hp->txt[2] && txt[3] == hp->txt[3] &&
       txt[4] == hp->txt[4] && txt[5] == hp->txt[5] &&
       txt[6] == hp->txt[6] && txt[7] == hp->txt[7]) {
      hp->created = now;
      return 0;
    }
  }

  /*
   * Not found -- allocate an element
   */

  hp = freelist->fwrd;
  if(hp == freelist) {
    /*
     * No free elements
     * at hand - allocate one
     */
    hp = ALLOC(hashent);
    hp->txt = Malloc(9);
    allocs++;
  } else {
    /*
     * Reuse expired element
     */
    unlk(hp);
  }

/*
  for(p1 = txt, p2 = hp->txt; c = *p1++;)
    *p2++ = c;
  *p2++ = 0;
*/
  p1 = txt; p2 = hp->txt;
  *p2++ = *p1++;  *p2++ = *p1++;
  *p2++ = *p1++;  *p2++ = *p1++;
  *p2++ = *p1++;  *p2++ = *p1++;
  *p2++ = *p1++;  *p2++ = *p1++;
  *p2++ = *p1++;

  hp->created = now++;
  insert_front(hashtable[hash], hp);

  return 1;
}

/*
 * Input a text file and make it into a word list
 */

word *read_txt_file(fp)
     FILE *fp;
{
  word *wp, *wl;
  char *p, *q;
  char buf[MAXLEN + 1];

  wl = (word*)NULL;
  while(fgets(buf, MAXLEN, fp)) {
    if(buf[0] == '#')
      continue;
    q = nextline(buf);
    *q = '\0';
    for(q = buf; *q && *q == ' '; q++);
    if(!*q)
      continue;
    p = save_str(buf);
    wp = ALLOC(word);
    wp->txt = p;
    wp->nextw = wl;
    wl = wp;
  }
  if(!feof(fp)) {
    fprintf(stderr, "failed to read dictionary: ");
    perror((char*)NULL);
  }
  (void)fclose(fp);
  return wl;
}

/*
 * Setup default rules - just
 * pass words through.
 */

word *setup_default_rules()
{
  word *wp = ALLOC(word);
  wp->txt = save_str(":");
  wp->nextw = (word*)NULL;
  return wp;
}

/*
 * dictionary list and rule list
 */

word *dictionary;
word *rules;

#define CLASS_v     0x00000001
#define CLASS_V     0x00000002
#define CLASS_c     0x00000004
#define CLASS_C     0x00000008
#define CLASS_w     0x00000010
#define CLASS_W     0x00000020
#define CLASS_p     0x00000040
#define CLASS_P     0x00000080
#define CLASS_s     0x00000100
#define CLASS_S     0x00000200
#define CLASS_l     0x00000400
#define CLASS_L     0x00000800
#define CLASS_u     0x00001000
#define CLASS_U     0x00002000
#define CLASS_d     0x00004000
#define CLASS_D     0x00008000
#define CLASS_a     0x00010000
#define CLASS_A     0x00020000
#define CLASS_x     0x00040000
#define CLASS_X     0x00080000
#define CLASS_quest 0x00100000

long attributes[256];
long char_to_attr[256];

void setup_tables()
{
  int a, c;

  for(c = 0; c < 256; c++) {
    a = 0;

    if(c == '?')
      a |= CLASS_quest;

    if(c == 'a' || c == 'e' || c == 'i' || c == 'o' || c == 'u') {
      a |= CLASS_v;
      a |= CLASS_C;
    } else {
      a |= CLASS_V;
      a |= CLASS_c;
    }

    if(isspace(c))
      a |= CLASS_w;
    else
      a |= CLASS_W;

    if(c == '.' || c == ',' || c == ':'  || c == ';' || c == '\'' ||
       c == '"' || c == '?' || c == '!'  || c == '`') {
      a |= CLASS_p;
    } else {
      a |= CLASS_P;
    }

    if(c == '$' || c == '%' || c == '^'  || c == '&' || c == '*' ||
       c == '(' || c == ')' || c == '-'  || c == '_' || c == '+' ||
       c == '=' || c == '|' || c == '\\' || c == '[' || c == ']' ||
       c == '{' || c == '}' || c == '#'  || c == '@' || c == '/' ||
       c == '~') {
      a |= CLASS_s;
    } else {
      a |= CLASS_S;
    }

    if(islower(c))
      a |= CLASS_l;
    else
      a |= CLASS_L;

    if(isupper(c))
      a |= CLASS_u;
    else
      a |= CLASS_U;

    if(isdigit(c))
      a |= CLASS_d;
    else
      a |= CLASS_D;

    if(isalnum(c))
      a |= CLASS_x;
    else
      a |= CLASS_X;

    if(isalpha(c))
      a |= CLASS_a;
    else
      a |= CLASS_A;

    attributes[c] = a;
  }

  bzero((void*)char_to_attr, sizeof(char_to_attr));
  char_to_attr['?'] = CLASS_quest;
  char_to_attr['v'] = CLASS_v;  char_to_attr['V'] = CLASS_V;
  char_to_attr['c'] = CLASS_c;  char_to_attr['C'] = CLASS_C;
  char_to_attr['w'] = CLASS_w;  char_to_attr['W'] = CLASS_W;
  char_to_attr['p'] = CLASS_p;  char_to_attr['P'] = CLASS_P;
  char_to_attr['s'] = CLASS_s;  char_to_attr['S'] = CLASS_S;
  char_to_attr['l'] = CLASS_l;  char_to_attr['L'] = CLASS_L;
  char_to_attr['u'] = CLASS_u;  char_to_attr['U'] = CLASS_U;
  char_to_attr['d'] = CLASS_d;  char_to_attr['D'] = CLASS_D;
  char_to_attr['a'] = CLASS_a;  char_to_attr['A'] = CLASS_A;
  char_to_attr['x'] = CLASS_x;  char_to_attr['X'] = CLASS_X;
}

/*
 * Checking of rules read
 */

void bad_rule(buf, c)
     char *buf, c;
{
  if(c == '\0')
    fprintf(stderr, "bad rule: %s (eol while reading)\n", buf);
  else
    fprintf(stderr, "bad rule: %s - offending character: %c\n", buf, c);
  exit(1);
}

void check_char(buf, pp)
     char *buf, **pp;
{
  char c = **pp; (*pp)++;
  if(c == '\0')
    bad_rule(buf, c);
}

void check_int(buf, pp)
     char *buf, **pp;
{
  char c = **pp; (*pp)++;
  if(c == '\0')
    bad_rule(buf, c);
  if(isdigit(c) || islower(c))
    return;
  bad_rule(buf, c);
}

void check_char_or_class(buf, pp)
     char *buf, **pp;
{
  char c = **pp; (*pp)++;
  switch(c) {
  case '\0':
    bad_rule(buf, c);
    break;
  case '?':
    check_char(buf, pp);
    break;
  }
}

void check_rules(rules)
     word *rules;
{
  word *rp;
  char *cp, c;
  for(rp = rules ; rp; rp = rp->nextw) {
    for(cp = rp->txt; c = *cp++; ) {
      switch(c) {

      case '<':
      case '>':
	check_int(rp->txt, &cp);
	break;

      case '^':
      case '$':
	check_char(rp->txt, &cp);
	break;

      case ':':
      case 'l':
      case 'u':
      case 'c':
      case 'r':
      case 'd':
      case 'f':
      case 'p':
	break;

      case 'o':
      case 'i':
	check_int(rp->txt, &cp);
	check_char(rp->txt, &cp);
	break;

      case 'x':
	check_int(rp->txt, &cp);
	check_int(rp->txt, &cp);
	break;

      case 's':
	check_char_or_class(rp->txt, &cp);
	check_char(rp->txt, &cp);
	break;

      case '@':
      case '!':
      case '/':
	check_char_or_class(rp->txt, &cp);
	break;

      case '=':
	check_int(rp->txt, &cp);
	check_char_or_class(rp->txt, &cp);
	break;

      default:
	bad_rule(rp->txt, c);
      }
    }
  }
}

/*
 * Line noise interpreter :-)
 */

void rcheck(wrd, rule)
     char *wrd, *rule;
{
  fprintf(stderr, "range check error in rule %s, word = %s\n", rule, wrd);
  exit(1);
}

#define PP_PASSON 1
#define PP_SKIP   2

int ascii_output = 0;

void interpret_rules(rules, dictionary, fp)
     word *rules;
     word *dictionary;
     FILE *fp;
{
  word *rp, *dp;
  char c, *cp;
  char *p1, *p2;
  char c1, c2;
  int done, todo, ok;
  int x, x1;
  long mask;

  char buf1[100], buf2[100];
  char *bufvec[2];
  int bufnum;
  char *buf, *nextbuf;
  int buflen, saved_buflen;

  bufvec[0] = buf1; bufvec[1] = buf2; bufnum = 0;
#define get_buf()   nextbuf = bufvec[bufnum++ % 2]
#define get_int(p)  (isdigit((*(p))) ? (*(p) - '0') : ((*(p)) - 'a' + 10))

  /*
   * Iterate over dictionary
   */
  for(dp = dictionary; dp; dp = dp->nextw) {
    /*
     * Iterate over rules
     */
    saved_buflen = strlen(dp->txt);
    for(rp = rules; rp; rp=rp->nextw) {
      /*printf("Processing rule: %s\n", rp->txt);*/

      /*
       * Interpret every dictionary word with
       * respect to current rewrite rule
       */
      done = 0; todo = PP_PASSON;
      buf = dp->txt; buflen = saved_buflen;

      for(cp = rp->txt; (c = *cp++) && !done; ) {
	/*printf("c = %c\n", c);*/
	switch(c) {
	case ':':                               /* No-op              */
	  break;

	case '<':                               /* Select short words */
 	  x = get_int(cp); cp++;
	  if(buflen >= x) {
	    todo = PP_SKIP;
	    done++;
	  }
	  break;

	case '>':                               /* Select long words   */
	  x = get_int(cp); cp++;
	  if(buflen <= x) {
	    todo = PP_SKIP;
	    done++;
	  }
	  break;

	case '^':                               /* Prepend character   */
	  c = *cp++;
	  get_buf();
	  p1 = buf; p2 = nextbuf;
	  *p2++ = c;
	  while(c = *p1++)
	    *p2++ = c;
	  *p2++ = '\0';
	  buf = nextbuf; buflen++;
	  break;

	case '$':                              /* Append character     */
	  get_buf();
	  p1 = buf; p2 = nextbuf;
	  while(c = *p1++)
	    *p2++ = c;
	  *p2++ = *cp++; *p2++ = '\0';
	  buf = nextbuf; buflen++;
	  break;

	case 'l':                              /* Lowercase word       */
	  get_buf();
	  for(p1 = buf, p2 = nextbuf; c = *p1++; )
	    *p2++ = isalpha(c) ? tolower(c) : c;
	  *p2++ = '\0';
	  buf = nextbuf;
	  break;

	case 'u':                             /* Uppercase word       */
	  get_buf();
	  for(p1 = buf, p2 = nextbuf; c = *p1++; )
	    *p2++ = isalpha(c) ? toupper(c) : c;
	  *p2++ = '\0';
	  buf = nextbuf;
	  break;

	case 'c':                            /* Capitalize word      */
	  get_buf();
	  p1 = buf; p2 = nextbuf;
	  c = *p1++;
	  *p2++ = isalpha(c) ? toupper(c) : c;
	  while(c = *p1++)
	    *p2++ = c;
	  buf = nextbuf;
	  break;

	case 'r':                            /* Reverse word         */
	  get_buf();
	  p1 = buf + buflen - 1; p2 = nextbuf;
	  while(p1 >= buf)
	    *p2++ = *p1--;
	  *p2++ = '\0';
	  buf = nextbuf;
	  break;

	case 'd':                           /* Duplicate word       */
	  get_buf();
	  p2 = nextbuf;
	  p1 = buf;
	  while(c = *p1++)
	    *p2++ = c;
	  p1 = buf;
	  while(c = *p1++)
	    *p2++ = c;
	  *p2++ = '\0';
	  buf = nextbuf; buflen *= 2;
	  break;

	case 'f':                           /* Reflect word         */
	  get_buf();
	  p1 = buf; p2 = nextbuf;
	  while(c = *p1++)
	    *p2++ = c;
	  p1 -= 2;
	  while(p1 >= buf)
	    *p2++ = *p1--;
	  *p2++ = '\0';
	  buf = nextbuf; buflen *= 2;
	  break;

	case 'p':                          /* Pluralize word        */
	  /*
	   * Warning: this is not done very carefully
	   */
	  get_buf();
	  p1 = buf + buflen - 1;
	  c = *p1; *p1 = '\0';
	  switch(c) {
	  case 'y':
	    (void)sprintf(nextbuf, "%sies", buf);
	    buflen += 2;
	    break;
	  case 'o':
	    (void)sprintf(nextbuf, "%soes", buf);
	    buflen += 2;
	    break;
	  case 's':
	    (void)sprintf(nextbuf, "%sses", buf);
	    buflen += 2;
	    break;
	  default:
	    (void)sprintf(nextbuf, "%s%cs", buf, c);
	    buflen++;
	    break;
	  }
	  *p1 = c;
	  buf = nextbuf;
	  break;

	case 'o':                          /* Overstrike character  */
	  get_buf();
	  x = get_int(cp); cp++;
	  if(x >= buflen)
	    rcheck(dp->txt, rp->txt);
	  p1 = buf; p2 = nextbuf;
	  while(c = *p1++)
	    *p2++ = c;
	  *p2++ = '\0';
	  buf = nextbuf;
	  buf[x] = *cp++;
	  break;

	case 'i':                          /* Insert character      */
	  x = get_int(cp); cp++;
	  c1 = *cp++;
	  get_buf();
	  p1 = buf; p2 = nextbuf;
	  while(x-- && (c = *p1++))
	    *p2++ = c;
	  *p2++ = c1;
	  while(c = *p1++)
	    *p2++ = c;
	  *p2++ = '\0';
	  buf = nextbuf;
	  buflen++;
	  break;

	case 'x':                          /* Extract substring    */
	  get_buf();
	  x = get_int(cp);  cp++;
	  x1 = get_int(cp); cp++;
	  if(x >= buflen)
	    rcheck(dp->txt, rp->txt);
	  for(p1 = buf+x, p2 = nextbuf, buflen = 0; x1--; ) {
	    *p2++ = *p1++;
	    buflen++;
	  }
	  *p2++ = '\0';
	  buf = nextbuf;
	  break;

	case 's':                          /* Replace (or swap)    */
	  get_buf();
	  if(*cp == '?') {
	    cp++;
	    c1 = *cp++; c2 = *cp++;
	    mask = char_to_attr[(int)c1];
	    for(p1 = buf, p2 = nextbuf; c = *p1++; ) {
	      if(attributes[(int)c] & mask)
		*p2++ = c2;
	      else
		*p2++ = c;
	    }
	    *p2++ = '\0';
	  } else {
	    c1 = *cp++; c2 = *cp++;
	    for(p1 = buf, p2 = nextbuf; c = *p1++; ) {
	      if(c == c1)
		*p2++ = c2;
	      else if(c == c2)
		*p2++ = c1;
	      else
		*p2++ = c;
	    }
	    *p2++ = '\0';
	  }
	  buf = nextbuf;
	  break;

	case '@':                          /* Purge                */
	  get_buf();
	  if(*cp == '?') {
	    cp++; c1 = *cp++;
	    mask = char_to_attr[(int)c1];
	    for(p1 = buf, p2 = nextbuf, buflen = 0; c = *p1++; ) {
	      if((attributes[(int)c] & mask) == 0) {
		*p2++ = c;
		buflen++;
	      }
	    }
	    *p2++ = '\0';
	  } else {
	    c1 = *cp++;
	    for(p1 = buf, p2 = nextbuf, buflen = 0; c = *p1++; ) {
	      if(c != c1) {
		*p2++ = c;
		buflen++;
	      }
	    }
	    *p2++ = '\0';
	  }
	  buf = nextbuf;
	  break;

	case '!':                          /* Patternmatch         */
	  if(*cp == '?') {
	    cp++; c1 = *cp++;
	    mask = char_to_attr[(int)c1];
	    for(p1 = buf; c = *p1; p1++) {
	      if(attributes[(int)c] & mask) {
		todo = PP_SKIP;
		done++;
		break;
	      }
	    }
	  } else {
	    c1 = *cp++;
	    for(p1 = buf; c = *p1; p1++) {
	      if(c == c1) {
		todo = PP_SKIP;
		done++;
		break;
	      }
	    }
	  }
	  break;

	case '/':                          /* Patternmatch         */
	  ok = 0;
	  if(*cp == '?') {
	    cp++; c1 = *cp++;
	    mask = char_to_attr[(int)c1];
	    for(p1 = buf; c = *p1; p1++) {
	      if(attributes[(int)c] & mask) {
		ok++;
		break;
	      }
	    }
	  } else {
	    c1 = *cp++;
	    for(p1 = buf; c = *p1; p1++) {
	      if(c == c1) {
		ok++;
		break;
	      }
	    }
	  }
	  if(!ok) {
	    todo = PP_SKIP;
	    done++;
	  }
	  break;

	case '=':                          /* Check character      */
	  x = get_int(cp); cp++;
	  if(x >= buflen)
	    rcheck(dp->txt, rp->txt);
	  if(*cp == '?') {
	    cp++; c1 = *cp++;
	    mask = char_to_attr[(int)c1];
	    if((attributes[(int)buf[x]] & mask) == 0) {
	      todo = PP_SKIP;
	      done++;
	    }
	  } else {
	    c1 = *cp++;
	    if(buf[x] != c1) {
	      todo = PP_SKIP;
	      done++;
	    }
	  }
	  break;

	default:
	  fprintf(stderr, "illegal character (%c) in rule %s\n", c, rp->txt);
	  exit(1);
	}
      }
      if(todo == PP_SKIP)
	continue;

      /*
       * Pad with zeroes
       */

      c = 1;
      for(x=0; x < 8; x++) {
	if(c)
	  c = buf[x];
	buf[x] = c;
      }
      buf[8] = '\0';

      if(is_uniq(buf)) {
	outwords++;
	if(ascii_output)
	  fprintf(fp, "%s\n", buf);
	else
	  fwrite(buf, 8, 1, fp);
      }
    }
  }
}

/*
 * Main program
 */

void usage()
{
  fprintf(stderr,
	  "usage: dpp [-a] [-r=rules] [-d=dictionary] [-o=output file]\n");
  exit(1);
}

void main(argc, argv)
  int argc;
  char **argv;
{
  char *dict_file, *rule_file, *output_file;
  int i;
  FILE *fpi, *fpo, *fpr;

  rule_file = dict_file = output_file = NULL;

    for( i=1;i<argc;i++ )
    {   if( argv[i][0]!='-' & argv[i][0]!='/' )
            usage();
//        if( argv[i][2]!='=' & argv[i][1]!='a' )
//            usage();
        switch( argv[i][1] )
        {
            case 'd':
                dict_file=argv[i]+3;
                break;
            case 'r':
                rule_file=argv[i]+3;
                break;
            case 'o':
                output_file=argv[i]+3;
                break;
            case 'a':
		      ascii_output++;
		      break;
            default:
                usage();
        }
    }

  if(dict_file) {
    if((fpi = fopen(dict_file, "r")) == NULL) {
      fprintf(stderr, "failed to open %s for reading: ", dict_file);
      perror(NULL);
      exit(1);
    }
  } else {
    fpi = stdin;
  }
  dictionary = read_txt_file(fpi);

  if(output_file) {
    if((fpo = fopen(output_file, "w")) == NULL) {
      fprintf(stderr, "failed to open %s for writing: ", output_file);
      perror(NULL);
      exit(1);
    }
  } else {
    fpo = stdout;
  }

  if(rule_file != (char*)NULL) {
    if((fpr = fopen(rule_file, "r")) == (FILE*)NULL) {
      fprintf(stderr, "failed to open %s for reading: ", rule_file);
      perror((char*)NULL);
      exit(1);
    }
    rules = read_txt_file(fpr);
  } else {
    rules = setup_default_rules();
  }

  setup_tables();
  check_rules(rules);
  interpret_rules(rules, dictionary, fpo);

  fprintf(stderr, "Reqs=%d Search=%d Alloc=%d Out=%d\n",
	  requests, searches, allocs, outwords);

  exit(0);
}

char *nextline(char *buf)
{	while( *buf!='\n' & *buf!=0 & *buf!=EOF)
		buf++;
	return( buf );
}

void bzero(void *buf, unsigned int len)
{	char *tmp;
	unsigned int i;

	tmp=(char *)buf;
	for( i=0;i<len;i++ )
		tmp[i]='\0';
}
